
############################################################
#  Refine the matrix Q by applying Newton's method to the
#  SOS (sum of squares) equation f = m'U'U m, where Q=U'U.
#
#  Input:
#      f        -  the polynomial
#      monos    -  the monomials
#      Q        -  the initial SOS matrix
#      eig_tol  -  the tolerance for eigenvalues
#      max_iter -  maximal number of iterations
#      tol      -  tolerance
#  Output:
#      the refinement of Q
#  Author: Bin LI  Date: 2008/1/19
#  Modify: Yue Ma  Date: 2010/12/03
############################################################
Newton_refine := proc(f, monos, Q, eig_tol, max_iter, tol)
    local mv,m,n,t,L,r,i,j,k,v,sL,F,W,b,A,tm,_q,tt, sL_update;

    
    m,n := LinearAlgebra[Dimension](Q);
    if m<>n or m<>LinearAlgebra[Dimension](monos) then
        error("Invalid input.");
    end if:
    # print(`|f-m'*Q*m|`,norm(expand(f-Vector[row](mv).Q.mv),2));

    tm := time();
    t,L := LinearAlgebra[Eigenvectors](Q); 
    r := 0;
    k := 0;
    v := [];
    sL := Matrix(m,m);
    tt:=LinearAlgebra[Dimension](t);
    for j from 1 to tt do
        L[1..-1,j] := L[1..-1,j] * sqrt(t[j]);
        if abs(t[j]) > eig_tol then
            r := r + 1;
            for i from 1 to m do
                if abs(L[i,j]) > 10^(-9) then
                    k := k + 1;
                    v :=[op(v), Re(L[i,j])];
                    sL[i,j] := _q[k];
                end if:
            end do:
        end if:
    end do:
    print(`the rank deficiency of Q `,m-r);
  

    v := convert(v,Vector);
    t := Vector[row](monos).sL;  
    F := Vector([coeffs(expand(f-add(i^2,i=t)),indets(monos))]); 
    W := VectorCalculus[Jacobian](F,[seq(_q[i],i=1..k)]); 
    b := eval(F,_q=v);
    t := LinearAlgebra[Norm](b,2);
    print(`|f- \sum f_i^2|`,t);
    print(`the dimension of Jacobian matrix`,LinearAlgebra[Dimension](W));
    for i from 1 to max_iter while t>tol do
        A := eval(W,_q=v);
        v := v - LinearAlgebra[LeastSquares](A,b,optimize);
        b := eval(F,_q=v);
        t := LinearAlgebra[Norm](b,2);
        print(`|f- \sum f_i^2|`,t);
    end do:
    print(cat(i-1,`  iterations `),time()-tm,`seconds`);
    sL := eval(sL,_q=v);
    sL_update:=sL; 
    
    for i from m to 1 by -1  do
      if LinearAlgebra[Norm](LinearAlgebra[Column](sL,i),2)=0 then
        sL_update:=LinearAlgebra[DeleteColumn](sL_update,i);
      fi:
    end do:

    return sL_update.LinearAlgebra[Transpose](sL_update);
end proc:


############################################################
#  Refine the matrix Q by applying Newton's method to the
#  SOS (sum of squares) equation f = m'U'U m, where Q=U'U.
#
#  Input:
#      f        -  the polynomial
#      monos    -  the monomials
#      Q        -  the initial SOS matrix
#      eig_tol  -  the tolerance for eigenvalues
#      max_iter -  maximal number of iterations
#      tol      -  tolerance
#  Output:
#      the refinement of Q
#  Author: Bin LI  Date: 2008/1/19
############################################################
Newton_refine_update := proc(f, monos::Vector, Q, eig_tol, max_iter, tol)
    local mv,m,t,L,r,i,j,k,v,sL,F,W,b,A,tm,_q,sL_update,digit_orgin,tt;

    m,t := LinearAlgebra[Dimension](Q); 
    if m<>t or m<>LinearAlgebra[Dimension](monos) then
        error("Invalid input.");
    end if:
    # print(`|f-m'*Q*m|`,norm(expand(f-Vector[row](mv).Q.mv),2));

    tm := time();
    t,L := LinearAlgebra[Eigenvectors](Q);
    r := 0;
    k := 0;
    v := [];
    sL := Matrix(m,m);
    tt:=LinearAlgebra[Dimension](t);
    for j from 1 to tt do
        L[1..-1,j] := L[1..-1,j] * sqrt(t[j]);
        if abs(t[j]) > eig_tol then
            r := r + 1;
            for i from 1 to m do
                if abs(L[i,j]) > 10^(-3) then
                    v:=[op(v), Re(L[i,j])];
                    k:=k+1;
                    sL[i,j] := _q[k];
                end if:
            end do:
        end if:
    end do:
    userinfo(1,Newton_refine_update,print(`the rank deficiency of Q `,m-r));
    # print(`|Q-L*L'|`,LinearAlgebra[Norm](Q-L.LinearAlgebra[Transpose](L)));
 
    v := convert(v,Vector); 
    t := Vector[row](monos).sL;
    F := Vector([coeffs(expand(f-add(i^2,i=t)),indets(monos))]); 
    W := VectorCalculus[Jacobian](F,[seq(_q[i],i=1..k)]); 
    b := eval(F,_q=v);   
    t := LinearAlgebra[Norm](b,2);
    userinfo(1,Newton_refine_update,`the orginal residue without Newton iteration:`, t);
    userinfo(1, Newton_refine_update,`the dimension of Jacobian matrix`,LinearAlgebra[Dimension](W));
    digit_orgin:=Digits;
    #Digits:=floor(Digits/2):
    for i from 1 to max_iter while t>tol do
        A := eval(W,_q=v);
        Digits:=floor(evalhf(Digits));
        v := v - LinearAlgebra[LeastSquares](A,b,optimize);
        Digits:=digit_orgin;
        #print(`Dinmension of A=`, LinearAlgebra:-Dimension(A));
        b := eval(F,_q=v);
        t := LinearAlgebra[Norm](b,2);
        userinfo(1,Newton_refine_update,`|f-\sum f_i^2|`, t);
    end do:
    userinfo(1,Newton_refine_update,print(cat(i-1,`  iterations `),time()-tm,`seconds`));
    userinfo(1,Newton_refine_update,`|f-\sum f_i^2|`, t);
    #Digits:=digit_orgin;
    sL := eval(sL,_q=v);
    sL_update:=sL; print(sL);
    for i from  LinearAlgebra[ColumnDimension](sL) to 1 by -1  do
      if LinearAlgebra[Norm](LinearAlgebra[Column](sL,i),2)=0 then
        sL_update:=LinearAlgebra[DeleteColumn](sL_update,i);
      fi:
    end do:
    
    return sL_update.LinearAlgebra[Transpose](sL_update);
  
end proc:


############################################################
#  Try to find rational SOS (sum of squares) decomposition
#  f-minv*g = \sum d_i*f_i^2 = m'P m by minimizing |P-Q|^2.
#
#  Input:
#      f       -  the numerator polynomial
#      g       -  the denominator polynomial
#      minval  -  the minimal value
#      monos   -  the monomials
#      Q       -  the approximate SOS matrix
#  Output:
#      the rational low bound of f/g
#      the coefficients in decomposition
#      the polynomials in decomposition
#  Author: Bin LI  Date: 2008/1/19
############################################################
rational_SOS := proc(f, g, minval, monos, Q)
    local mv,m,t,rf,rg,rm,rQ,A,b,d,i,j,k,_r,_p,_q,tm;

    mv := Vector(monos);
    m,t := LinearAlgebra[Dimension](Q);
    if m<>t or m<>LinearAlgebra[Dimension](mv) then
        error("Invalid input.");
    end if:
    # t := f - minval*g - Vector[row](mv).Q.mv;
    # print(`|f-r*g-m'*Q*m|`,norm(expand(t),2));

    tm := time();
    rf := convert(f,rational,exact);
    rg := convert(g,rational,exact);
    rm := convert(minval,rational,exact);
    rQ := map(i->convert(i,rational,exact),Q);

    t := rf - rm*rg - Vector[row](mv).rQ.mv;
    t := t*_r + add(_p[i]*mv[i],i=1..m)*add(_q[i]*mv[i],i=1..m);
    t := [coeffs(expand(t),indets(mv))];
    b := map(i->coeff(i,_r),t);
    t := [seq(t[i]-b[i]*_r,i=1..nops(t))];
    t := add(b[i]*t[i]/norm(t[i],2)^2,i=1..nops(t));
    t := [seq(coeff(t,_p[i]),i=1..m)];
    A,b := LinearAlgebra[GenerateMatrix](t,[seq(_q[i],i=1..m)]);
    rQ := rQ + A;
    print(`the distance between Q and rQ is `,
          evalf(LinearAlgebra[Norm](A,Frobenius)));

    # t := rf - rm*rg - Vector[row](mv).rQ.mv;
    # print(`|f-r*g-m'*rQ*m|`,norm(expand(t),2));
    rQ := (LinearAlgebra[Transpose](rQ)+rQ)/2;
    d := Vector(m);
    for i from 1 to m do
        t := rQ[i,i];
        if t > 0 then
            d[i] := 1/t;
            for j from i+1 to m do
                for k from j to m do
                    rQ[j,k] := rQ[j,k] - rQ[i,j]*rQ[i,k]/t;
                end do:
            end do:
        elif LinearAlgebra[Norm](rQ[i,i..-1])>0 then
            print(`rQ is not positive semidefinite.`);
            return [],[],[];
        end if:
        rQ[i+1..-1,i] := 0;
    end do:
    print(rm,` is the lower bound.`);
    print(time()-tm,` seconds`);

    return rm,d,rQ.mv;
end proc:


###############################################################################################################
# Try to judge whether Q is positive semidefinite and give a decomposition Q=Transpose(L).D.L 
# Input:
#         Q   ---   matrix
# Outputs:
#         d   ---   vector
#         rQ  ---   upper triangular matrix
###############################################################################################################

Judge_exact_PSD:=proc(Q)

   local m,t,rQ,d,i,j,k;

    m:= LinearAlgebra[RowDimension](Q);
       
    rQ := map(i->convert(i,rational,exact),Q);
    rQ := (LinearAlgebra[Transpose](rQ)+rQ)/2;
    d := Vector(m);
    for i from 1 to m do
        t := rQ[i,i];
        if t > 0 then
            d[i] := 1/t;
            for j from i+1 to m do
                for k from j to m do
                    rQ[j,k] := rQ[j,k] - rQ[i,j]*rQ[i,k]/t;
                end do:
            end do:
        elif LinearAlgebra[Norm](rQ[i,i..-1])>0 then
            print(`rQ is not positive semidefinite.`);
            return [],[];
        end if:
        rQ[i+1..-1,i] := 0;
    end do:
     
    print(`rQ is positive semidefinite.`);
    return d,rQ;
end proc:





cutoff_matrix:=proc(M,tol)
  local i,n,m,j,res_M,temp_M,temp_number;
   m,n:=LinearAlgebra:-Dimension(M);
   res_M:=Matrix(m,n);
   #temp_M:=map(x->convert(x,rational,tol),M);
   temp_M:=M;
   for i from 1 to m do
     for j from 1 to n do
       temp_number:=temp_M[i,j]:
      if abs(evalf(temp_number))>10^(-tol) then
        res_M[i,j]:=convert(evalf(temp_number),rational,tol);
      fi;
     end do:
   end do:
   return res_M;
 end proc;
    
