function [X1,iter] = MFPC(n,r,A,num,b,maxinneriter,tol,eta,continuation)
%%**********************************************************************
%%  This is a Matlab implementation of modified fixed point continuation 
%%  method (MFPC) for solving the nuclear norm regularized semidefinite 
%%  least squares problem: 
%%
%%  min_X {mu ||X||_* + 1/2 || A(X) - b ||_2^2}
%%  where X is a positive semidefinite matrix. 
%%
%%  Note:  We conpute the full Schur decomposition at each iteration.
%%  
%% Input: 
%%        n            = size of the matrix.
%%        r            = the rank of the matrix.
%%        A,num        = function handle to evaluation A(X).
%%        maxinneriter = maximum number of iterations per inner loop.
%%        eta          = rational number <1 to determine how much to decrease
%%                       mu at every step. e.g. eta=1/4.
%%        continuation = 1/0 to determine whether to use continuation technique
%%                       (i.e. a series of mu, aka the "outer loop", until mu = mu_final). 
%%
%% Output: 
%%        X    = the numerical Gram matrix.
%%        iter = the number of iterations over all inner and outer loops. 
%%         
%% 
%% Reference:
%%   "The Minimum-Rank Gram Matrix Completion via Modified Fixed Point
%%     Continuation Method."
%%   Yue Ma and Lihong Zhi
%%
%%  code by Yue Ma   yma@mmrc.iss.ac.cn, November 2010
%% ********************************************************************** 
   p = length(b);
   sr = p/n^2;
   fr = r*(2*n-r+1)/2/p;
   fprintf('sr:%f\n',sr);
   fprintf('fr:%f\n',fr);
   
   if fr< 0.4 || fr >= 0.4
       fprintf('This is a "hard" problem! \n');
   else
       fprintf('This is an "easy" problem! \n');
   end

   time1 = cputime;
   
   maxiter = 1e+5; % maximum iteration number for the outer loop
   normb = norm(b);
%% initial mu and target mu
%%   
   temp = Adj_A(A,num,b,n);
   if n < 500
           nrm2Adjb = norm(temp);
   else
           T = laneig(temp,3); 
           nrm2Adjb = T(1);
   end
   mu_final = (1e-4)*nrm2Adjb;
   
   if continuation
       mu = eta*nrm2Adjb;
   else 
       mu = mu_final;
   end
   
%% estimate Lipschitz constant = largest eigenvalue of MAP_A(Adj_A). 
%%
      options.tol   = 1e-6; 
      options.issym = true; 
      options.disp  = 0; 
      options.v0    = randn(p,1);
      Lipschitz_const = eigs(@(y)Map_A(A,num,Adj_A(A,num,y,n)),p,1,'LM',options); 
      tau = 2/full(Lipschitz_const);  
%% Initialization
%%   
         
   X1 = zeros(n,n);
   numin = 0;

   
%% main loop
%%
   fprintf('MFPC:\n');
   %fprintf('Iteration begins:\n');
   %fprintf('*******************************************************\n');
   
   for iter = 1:maxiter
          if mu == mu_final
                 maxinneriter = 5000;
          end
          
          if iter == 1
              Y = tau*temp;
          else
              g = Adj_A(A,num,AX1_b,n);
              Y = X1-tau*g;  
          end
          X0 = X1;
      
          Y = (Y+Y')/2;
          [U, T] = schur(Y); 
                              
          nu = tau*mu; 
          sigma = diag(T); 
          ind = find(sigma > nu); r = length(ind);         
          Te = diag(sigma(ind)-nu);
          Ue = U(:,ind);
                        
          X1 = Ue*Te*Ue';
          rX = X1-X0;
          norm_new = norm(rX,'fro');         
          normX0 = max(norm(X0,'fro'),1);
          residualX = norm_new/normX0;
         
          
          AX1_b = Map_A(A,num,X1)-b;
          normAX1_b = norm(AX1_b);   
          residualA = normAX1_b/normb;  
          
          numin = numin +1;
          %fprintf('***************************\n');
          %fprintf('Iteration:%d\n',numin);
          %fprintf('Rank:%d\n',r);
          %fprintf('Residual_X:%6.8e\n',residualX);
          %fprintf('The relative error of AX-b is %6.8e\n',residualA);
          %fprintf('The nuclear norm is %6.8f\n',norm(diag(Te),1));
          %no need to compute in large scale examples
               
                
          if residualA > 1e5
             disp('Divergence!');
              break;
          end
  
          
          if residualA < tol || numin >= maxinneriter
              numin = 0;
              %fprintf('*******************************************************\n');
              if mu == mu_final
                  eTime = cputime - time1; 
                  fprintf('Rank:%d\n',r);
                  fprintf('Residual_X:%6.8e\n',residualX);
                  fprintf('The relative error of AX-b is %6.8e\n',residualA);
                  fprintf('The number of iteration is %d\n',iter);
                  fprintf('The computing time is %f\n\n',eTime);
                  return
              end

              mu = max(mu*eta,mu_final);
          end  
   end
%%

%-------------------------------------------------------------------------%
%                            Subfunctions
%-------------------------------------------------------------------------%

   function y = Map_A(subA,num,X)
%   y = A(X)    
       p = length(num);
       y = zeros(p,1);
       t = 0;
       for i = 1:p 
           k = num(i);
           if k == 1
               y(i) = X(subA(t+1,1),subA(t+1,2));
           else
               z=zeros(k,1);
               for j = 1:k
                   z(j) = X(subA(t+j,1),subA(t+j,2));
               end
               y(i) = sum(z);            
           end
           t = t+k;
       end
       
       
%-------------------------------------------------------------------------%

function [X] = Adj_A(subA,num,y,n)
%    X=A^*(y)
      p = length(y);
      X = zeros(n,n);
      t = 0;
      for i = 1:p
          k = num(i);
          if y(i)~=0             
              if k==1
                  X(subA(t+1,1),subA(t+1,2)) = y(i);
              else
                  for j = 1:k
                      X(subA(t+j,1),subA(t+j,2)) = y(i);
                  end 
              end
          end
          t = t+k; 
      end
%%
